/* GStreamer H264 video encoder
 * Copyright (C) 2010 Miguel Verdu <miguel.verdu@nokia.com>
 *
 * h264encoder.c: Common interface for H264 encoders
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * 
 * SECTION:h264encoderif
 * @title: H264 Encoder Interface
 * @short_description: Interface to h264 video encoders
 * 
 * This interface provides a minimum set of properties that H264 video encoders
 * should implement. It is targetted mainly for HW accelerated encoders
 * <ulink url="http://www.itu.int/rec/T-REC-H.264-201003-I/">H.264 is a ITU</ulink>
 * specified video coding standard.
 *
 * GStreamer conventions do a pretty good job encapsulating the behaviour of 
 * decoders in general, therefore video decoders tend to have few properties if
 * any. And the few they have are related to debugging, rendering or QoS.
 *
 * But for encoders the situation is quite different. With encoders the number
 * of possible parameters is up to the encoder developer For example: x264 has
 * around 110+ parameters. Only 30 get exposed at the Gstreamer level. 
 * This is an extreme case, but shows the need for more control needed in the
 * encoder side.
 *
 * The main objective of this interface is to achieve some degree of uniformity
 * in the name of the properties. The implementation of the interface does not
 * mandate that the element implementing will support all the features, but if it 
 * does they will be accessed as the interface defines.
 *
 * <emphasis>Events:</emphasis>
 * It is not possible to implement support to handle certain events in an interface
 * (at least I don't know how to do it), but elements implementing this interface
 * are encouraged to handle the following event.
 *
 * GSTForceKeyUnit with support for the following fields:
 * <itemizedlist>
 * <listitem><para>first-macroblock</para></listitem>
 * <listitem><para>macroblock-count</para></listitem>
 * <listitem><para>picture-id</para></listitem>
 * </itemizedlist>
 *
 * The purpose of this event is to be able to force the insertion of an I frame
 * or an I slice in case of detail information about the corruped  MB's is provided
 *
 * <emphasis>Presets:</emphasis>
 * There is a usual feature found in video encoder interfaces that is missing
 * from this interface, this is presets.
 * Presets should be implemented by using the #GstPreset interface defined in GStreamer
 * This interface actually has a default implementation that is enough for most use cases,
 * so the cost for element developers is minimal. 
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "meegoh264encoder.h"


/*
 * Defaults values for properties 
 *
 */
#define ARG_PROFILE_DEFAULT            		1       /* 'Base Profile'  */
#define ARG_RC_MODE_DEFAULT					MEEGO_GST_H264_ENCODER_RATE_CONTROL_VARIABLE
#define ARG_BITRATE_DEFAULT					0       /*Let the encoder choose */
#define ARG_MAX_BITRATE_DEFAULT				0       /*Let the encoder choose */
#define ARG_MIN_QUANTIZER_DEFAULT			-1
#define ARG_MAX_QUANTIZER_DEFAULT			-1
#define ARG_INTRA_QUANTIZER_DEFAULT			-1
#define ARG_INTER_QUANTIZER_DEFAULT			-1
#define ARG_DELAY_DEFAULT					200     /*ms */
#define ARG_TOLERANCE_DEFAULT				0.15    /*15 % */
#define ARG_KEYFRAME_PERIOD_DEFAULT			1		/* 1 second */
#define ARG_SLICE_SIZE_DEFAULT				0
#define ARG_INTRA_REFRESH_MODE_DEFAULT		MEEGO_GST_H264_ENCODER_INTRA_REFRESH_NONE
#define ARG_INTRA_REFRESH_RATE_DEFAULT		3.0


static void meego_gst_h264_encoder_iface_base_init (MeegoGstH264EncoderClass *
    klass);
static void meego_gst_h264_encoder_iface_class_init (MeegoGstH264EncoderClass *
    klass);



GType
meego_gst_h264_encoder_get_type (void)
{
  static GType meego_gst_h264_encoder_type = 0;

  if (!meego_gst_h264_encoder_type) {
    static const GTypeInfo meego_gst_h264_encoder_info = {
      sizeof (MeegoGstH264EncoderClass),
      (GBaseInitFunc) meego_gst_h264_encoder_iface_base_init,   /* base_init */
      NULL,                     /* base_finalize */
      (GClassInitFunc) meego_gst_h264_encoder_iface_class_init, /* class_init */
      NULL,                     /* class_finalize */
      NULL,                     /* class_data */
      0,
      0,                        /* n_preallocs */
      NULL,                     /* instance_init */
    };

    meego_gst_h264_encoder_type = g_type_register_static (G_TYPE_INTERFACE,
        "MeegoGstH264Encoder", &meego_gst_h264_encoder_info, 0);
    g_type_interface_add_prerequisite (meego_gst_h264_encoder_type,
        GST_TYPE_IMPLEMENTS_INTERFACE);
  }

  return meego_gst_h264_encoder_type;
}

static void
meego_gst_h264_encoder_iface_base_init (MeegoGstH264EncoderClass * klass)
{



}



static void
meego_gst_h264_encoder_iface_class_init (MeegoGstH264EncoderClass * klass)
{

   /**
	*  MeegoGstH264Encoder:profile
	*
	*  Profile to be used for encoding as defined by the standard 
	* 
	*/
  g_object_interface_install_property (klass,
      g_param_spec_enum (MEEGO_GST_H264_ENCODER_PROP_PROFILE, "H.264 profile",
          "Apply restrictions to meet H.264 Profile constraints. This will "
          "override other properties if necessary.",
          GST_TYPE_MEEGO_H264_ENCODER_PROFILE, ARG_PROFILE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:rate-control-mode
	*
	* Rate control mode
	* This parameter controls the type of bitrate control algorithm the encoder will use
	* The interface defines several modes in #MeegoGstH264EncoderRateControlMode
	* 
	*/
  g_object_interface_install_property (klass,
      g_param_spec_enum (MEEGO_GST_H264_ENCODER_PROP_RATE_CONTROL_MODE,
          "Rate control mode", "Bitrate control algorithm used ",
          GST_TYPE_MEEGO_H264_ENCODER_RATE_CONTROL_MODE, ARG_RC_MODE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:bit-rate
	*
	* Bitrate in bits per second 
	* This property can be changed during the encoding process as opposed to
	* #MeegoGstH264Encoder:max-bitrate that is set only before the encoding begins
	* If set to 0 (default) encoder will choose a value based on frame size and
	* frame rate that fits in a valid standard level.
	*/
  g_object_interface_install_property (klass,
      g_param_spec_uint (MEEGO_GST_H264_ENCODER_PROP_BITRATE,
          "bitrate", "Bitrate in  bits per second"
          "This property can be changed during the encoding process",
          0, G_MAXUINT32, ARG_BITRATE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));


   /**
	*  MeegoGstH264Encoder:max-bit-rate
	*
	* Maximum Bitrate in bits per second 
	* This property can NOT be changed during the encoding process as opposed to
	* #MeegoGstH264Encoder:bitrate 
	* It is used as a hint to the encoder to allow for a change of bitrate 
	* during the encoding process. The encoder will not allow (or ignore) a change
	* of bitrate to a value bigger than the one set by this property 
	* If set to 0 (default) encoder will choose a value based on frame size and
	* frame rate that fits in a valid standard level.
	*/
  g_object_interface_install_property (klass,
      g_param_spec_uint (MEEGO_GST_H264_ENCODER_PROP_MAX_BITRATE,
          "Max bitrate", "Maximum Bitrate in  bits per second."
          " This property can NOT be changed during the encoding process\n"
          "\t\t\tIt is used as a hint to the encoder to allow for a change of bitrate\n"
          "\t\t\tduring the encoding process. The encoder will not allow (or ignore) a change\n"
          "\t\t\tof bitrate to a value bigger than the one set by this property",
          0, G_MAXUINT32, ARG_MAX_BITRATE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

  // In case of RC-mode = None

   /**
	*  MeegoGstH264Encoder:min-qp
	*
	* Minimum Quantization Parameter to be used during the encoding process
	* The value -1 is used as default value, so that the encoder decides
	* which value is more appropiate
	*/
  g_object_interface_install_property (klass,
      g_param_spec_int (MEEGO_GST_H264_ENCODER_PROP_MIN_QP,
          "Minimum QP ", "Minimum QP value to be used",
          -1, 51, ARG_MIN_QUANTIZER_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:max-qp
	*
	* Maximum Quantization Parameter to be used during the encoding process	
	* The value -1 is used as default value, so that the encoder decides
	* which value is more appropiate
	*/
  g_object_interface_install_property (klass,
      g_param_spec_int (MEEGO_GST_H264_ENCODER_PROP_MAX_QP,
          "Maximum QP ", "Maximum QP value to be used",
          -1, 51, ARG_MAX_QUANTIZER_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:intra-qp
	*
	* Quantization Parameter to be used during the encoding process of I frames
	* When used with a rate control algorithm diferent than "none" this parameter
	* is used only for the first I frame.
	*/
  g_object_interface_install_property (klass,
      g_param_spec_int (MEEGO_GST_H264_ENCODER_PROP_INTRA_QP,
          "Intra QP ", "QP value to be used for Intra frames",
          -1, 51, ARG_INTRA_QUANTIZER_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:inter-qp
	*
	* Quantization Parameter to be used during the encoding process of inter 
	* predicted frames (P and B).
	*/
  g_object_interface_install_property (klass,
      g_param_spec_int (MEEGO_GST_H264_ENCODER_PROP_INTER_QP,
          "Inter QP ", "QP value to be used for Inter predicted frames",
          -1, 51, ARG_INTER_QUANTIZER_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));


   /**
	*  MeegoGstH264Encoder:delay
	*
	* Maximum delay in miliseconds that the encoder will produce.
	* The delay is nothing else than a parameter for a simplified VBV model
	* This property is only used in case the rc-mode is Constant Bitrate mode. 
	* The buffer size used in this simplified VBV is equal to :
	*        (delay) x (bitrate) 
	*/
  g_object_interface_install_property (klass,
      g_param_spec_int (MEEGO_GST_H264_ENCODER_PROP_DELAY,
          "Delay in ms ", "Maximum Delay allowed in Constant Bitrate mode",
          0, G_MAXINT, ARG_DELAY_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:rate-tol
	*
	* Maximum bitrate tolerance allowed
	* This property is only used in case the rc-mode is Variable Bitrate mode. 
	* The bitrate of the encoded sequence should be bounded in the region
	*  [ b*(1-tol), b*(1+tol)]
	* 
	* where:
	*  b   is the target bitrate specified in the property bitrate
	*  tol is the allowed tolerance as percentage
	*
	*   
	*/
  g_object_interface_install_property (klass,
      g_param_spec_float (MEEGO_GST_H264_ENCODER_PROP_RATE_TOLERANCE,
          "Bitrate tolerance in %", "Percentage of target bitrate that "
          "the encoder is allowed to deviate from",
          0.0, 1.0, ARG_TOLERANCE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));


   /**
	*  MeegoGstH264Encoder:keyframe-period
	*
	* Number of seconds between two I frames
	* The actual number of frames between I-frames is keyframe-period * framerate
	* Every time this property is set an I frame is inserted and the cycle
	* starts. In this way an application can force the insertion of an I frames
	* A value of 0 will insert a single I frame at the begining of the stream.
	*
	* Other methods of inserting I frames on demand is by sending  the event
	* GstForceKeyUnit
	*/
  g_object_interface_install_property (klass,
      g_param_spec_uint (MEEGO_GST_H264_ENCODER_PROP_KEY_FRAME_PERIOD,
          "I-frame interval", "Number of seconds between two I frames",
          0, G_MAXUINT, ARG_KEYFRAME_PERIOD_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));

   /**
	*  MeegoGstH264Encoder:slice-size-mb
	*
	* Number of MacroBlocks in a single NAL Unit
	* A value of 0 is used as default and equal to one slice per frame
	* i.e:
	*   1 NAL Unit = 1 Access Unit 
	*
	* Application should estimate the size in bytes of the NAL Unit based one
	* the target bitrate and rate control algorithm.
	* (*NOTE: Maybe some encoders can implement also a slice-size-byte )
	*/
  g_object_interface_install_property (klass,
      g_param_spec_uint (MEEGO_GST_H264_ENCODER_PROP_SLICE_SIZE_MB,
          "Number of MB's per slice",
          "Number of MacroBlocks in a slice (NAL unit)", 0, G_MAXUINT,
          ARG_SLICE_SIZE_DEFAULT, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));


   /**
	*  MeegoGstH264Encoder:intra-refresh
	*
	* Method for Intra-MB refresh on Inter predicted frames.
	* Different strategies can be followed, is up to the implementation 
	* decide which one it supports.
	* Two of the most common are:
	* 	- Adaptative Intra Refresh : AIR 
	*   - Cyclic Intra Refresh: CIR 	 
    *
	*  see #MeegoGstH264EncoderIntraRefreshMode for more details
	*
	*/
  g_object_interface_install_property (klass,
      g_param_spec_enum (MEEGO_GST_H264_ENCODER_PROP_INTRA_REFRESH_MODE,
          "Intra refresh mode",
          "Modes supported for I-MB insertion on P frames",
          GST_TYPE_MEEGO_H264_ENCODER_INTRA_REFRESH_MODE,
          ARG_INTRA_REFRESH_MODE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));


   /**
	*  MeegoGstH264Encoder:intra-refresh-rate
	*
	* Number of seconds it takes the intra refresh method to refresh the whole
	* frame.
	* Only meaningful in cyclic intra refresh modes or similar. It depends on 
	* the size of the kernel/pattern used to refresh the image.
	* 
	*/
  g_object_interface_install_property (klass,
      g_param_spec_float (MEEGO_GST_H264_ENCODER_PROP_INTRA_REFRESH_RATE,
          "Frames until whole frame is refreshed ", "Number of frames until"
          " all MB's in a frame have been encoded as I-MB at least once\n"
          "\t\t\tOnly useful in intra refresh modes like cyclic",
          0, G_MAXUINT, ARG_INTRA_REFRESH_RATE_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));




}
