/* GStreamer Video Render switch
 * Copyright (C) 2010 Sami Kyöstilä <sami.kyostila@nokia.com>
 * Copyright (C) 2010 Miguel Verdu <miguel.verdu@nokia.com>
 *
 *  videorenderswitch.c: video render mode switch interface design
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * SECTION:meegogstvideorenderswitch
 * @short_description: Interface to switch the rendering mode of video frames
 *
 * This document describes a high-level mechanism that enables efficient switching
 * of rendering mode of video frames.
 * This interface is inteded for sinks implementing more than one rendering interface,
 * at the moment only two interfaces are available:
 * <itemizedlist>
 * <listitem><para>
 *  X overlay: to render video on an X-drawable.
 * </para></listitem>
 * <listitem><para>
 *  Video Texture: to provie video frames in a GL texture. Render operation is 
 *  controlled by the application.
 * </para></listitem>
 * </itemizedlist>
 *
 * Elements implementing this interface also implement VideoTexture and X-overlay
 * interfaces at least. More could be possible. But those are the ones supported
 * at the moment. The properties in this interface are intentionaly defined as 
 * strings and not enums to allow other rendering modes to be added in the future
 * whithout changes in the interface.
 *
 * The rational of this interface is to provide a standarize way to switch between
 * those two rendering modes. The need for the switch is power efficiency. 
 * HW overlays tend to be more efficient than Textured video that goes through
 * the GPU
 *
 * Usually the sink implementing this interface has property to give access 
 * to the X-Display structure even though this is not defined by the X-overlay
 * interface.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include "meegovideorenderswitch.h"



static void meego_gst_video_renderswitch_iface_class_init (MeegoGstVideoRenderSwitchClass *
    klass);

static gboolean meego_gst_video_renderswitch_mode_check (GList * supported,
    const gchar * mode);

GType
meego_gst_video_renderswitch_get_type (void)
{
  static GType meego_gst_video_renderswitch_type = 0;

  if (!meego_gst_video_renderswitch_type) {
    static const GTypeInfo meego_gst_video_renderswitch_info = {
      sizeof (MeegoGstVideoRenderSwitchClass),
      NULL,                     /* base_init */
      NULL,                     /* base_finalize */
      (GClassInitFunc) meego_gst_video_renderswitch_iface_class_init, /* class_init */
      NULL,                     /* class_finalize */
      NULL,                     /* class_data */
      0,
      0,                        /* n_preallocs */
      NULL,                     /* instance_init */
    };

    meego_gst_video_renderswitch_type = g_type_register_static (G_TYPE_INTERFACE,
        "MeegoGstVideoRenderSwitch", &meego_gst_video_renderswitch_info, 0);
    g_type_interface_add_prerequisite (meego_gst_video_renderswitch_type,
        GST_TYPE_IMPLEMENTS_INTERFACE);
  }

  return meego_gst_video_renderswitch_type;
}



static void
meego_gst_video_renderswitch_iface_class_init (MeegoGstVideoRenderSwitchClass * klass)
{

  /**
   *  MeegoGstVideoRenderSwitch:modes-supported
   *
   * property to query the supported rendering modes available
   * 
   */
  g_object_interface_install_property (klass,
      g_param_spec_pointer ("modes-supported",
          "List of supported modes ",
          "GList of strings with the possible values that can be passed to the property render-mode",
          G_PARAM_READABLE));

  /**
   *  MeegoGstVideoRenderSwitch:render-mode
   *
   *  Change the render mode
   * 
   */
  g_object_interface_install_property (klass,
      g_param_spec_string ("render-mode",
          "Video render mode",
          "Select the current video render mode",
          VIDEO_RENDERSWITCH_XOVERLAY_MODE, G_PARAM_READWRITE));

  klass->is_mode_supported = meego_gst_video_renderswitch_mode_check;

}


static gboolean
meego_gst_video_renderswitch_mode_check (GList * supported, const gchar * mode)
{
  gboolean ret = FALSE;

  if (supported) {
    if (g_list_find_custom (supported, mode, (GCompareFunc) strcmp))
      ret = TRUE;

  }
  return ret;
}
