/* GStreamer Video Texture Streaming
 * Copyright (C) 2010 Sami Kyöstilä <sami.kyostila@nokia.com>
 * Copyright (C) 2010 Miguel Verdu <miguel.verdu@nokia.com>
 *
 * videotexture.c: Video Texture streaming interface design
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
/**
 * SECTION:videotexture
 * @short_description: Interface to offer graphic textures with video content
 * by elements that supporting it
 * 
 * This document describes a high-level mechanism that enables efficient streaming
 * of video frames from a GStreamer pipeline to OpenGL ES 2.0 textures. The main
 * features of this design are:
 * <itemizedlist>
 * <listitem><para>
 * Zero-copy streaming -- No intermediate buffer copies are introduced to the
 * pipeline by the streaming mechanism.
 * </para></listitem>
 * <listitem><para>
 * Multi-vendor proof -- Applications are not tied to GPU-vendor specific APIs.
 * Functionality can be implemented by the element using diferent GL extensions
 * </para></listitem>
 * <listitem><para>
 * Cross-process streaming -- The producer and consumer of video frames can
 * reside in distinct OS processes.
 * </para></listitem>
 * </itemizedlist>
 *
 * The document first presents the high-level, cross-platform streaming interface.
 * This is followed by an sample call sequence showing how the streaming is
 * controlled.
 *
 * <refsect2>
 * <title>2.0 GStreamer texture sink </title>
 * <para>
 * |[
 *
 *   GST pipe
 *                                Frame ready signal   Render
 *       |        +---------+      +-------------+      +----+
 *       +------> | GST tex | ---> | Application | ---> | GL |
 *                | sink    | <--- |             | <--- |    |
 *                +---------+      +-------------+      +----+
 *                             Acquire frame        Sync
 *                             Bind/unbind frame
 *                             Release frame
 *
 *
 * ]|
 * 
 * A new GStreamer sink element is introduced to provide a hardware-neutral API
 * for controlling texture streaming. While this sink provides a uniform interface
 * for applications, its internal implementation is expected vary from one
 * platform to another.
 * In this situation AV sync is out of control of the sink and becomes responsibility
 * of the application
 * 
 * The texture sink exports the following property to the user:
 * 
 *  EGL-display         : The application provided EGL display which will be used
 *                        rendering.
 * 
 * 
 * Additionally, the following signals are provided:
 * 
 *  "frame-ready"       : void user_function (GstElement* object,
 *                                            guint frame,
 *                                            gpointer user_data);
 *
 *                        Triggered when a new frame is available from the sink
 *                        element. frame specifies an opaque, non-zero
 *                       identifier for the latest available frame. The
 *                        application should call "acquire-frame" as a response
 *                       to this signal.
 * </para>
 * </refsect2>                      
 * <refsect2>
 * <title> Single process usage sequence  </title>
 * <para>
 * The following sequence shows how the texture streaming sink is used by an 
 * application where video decoding  and playback happens in a single process.
 *
 * Initialization:
 *
 *  1. The application creates a GStreamer pipeline which terminates at a
 *     texture sink.
 *  2. The application connects the frame-ready signal appropriately.
 *  3. The application compiles a program with a fragment shader such as
 *     the following:
 *
 * |[
 *        #version 100
 *        #extension GL_OES_EGL_image_external: enable
 *        uniform samplerExternalOES video;
 *        varying mediump vec2 texCoord;
 *
 *        void main()
 *        {
 *            gl_FragColor = texture2D(video, texCoord);
 *        }
 * ]|
 *
 *Playback:
 *
 *  1. The frame-ready signal is triggered by the sink implementing the interface.
 *  2. The application calls gst_video_texture_acquire_frame with the frame
 *     identifier given by the previous signal.
 *  3. The application calls gst_video_texture_bind_frame(sink,GL_TEXTURE_EXTERNAL_OES, frame).
 *  4. The application uses OpenGL (ES) to draw the video frame.
 *  5. The application calls gst_video_texture_bind_frame(sink,GL_TEXTURE_EXTERNAL_OES, -1).
 *  6. The application places an EGL fence sync to the command stream by
 *     calling
 *
 *         EGLSyncKHR sync = eglCreateSyncKHR(dpy, EGL_SYNC_FENCE_KHR, NULL);
 *
 *  7. The application calls gst_video_texture_release_frame(sink,frame, sync).
 *  8. The application calls eglSwapBuffers() to display the graphics on the
 *     screen.
 * </para>
 * </refsect2>                      
 * <refsect2>
 * <title> Multiple process usage sequence  </title>
 * <para>
 * In a multi process streaming scenario, the video source and graphics rendering 
 * are split into separate OS processes. To enable this, the texture stream 
 * provides a way to transport video buffer handles across process boundaries. 
 * Note that the flow control of the buffers must still be managed by the 
 * application by calling the appropriate functions of the sink.
 * 
 * |[
 *
 *   GST pipe                                    App IPC
 *                            Frame ready signal   | 				Render
 *       |     +---------+      +-------------+    |    +--------+        +----+
 *       +---> | GST tex | ---> | Application | ---|--> |  App   |------->| GL |
 *             | sink    | <--- |   dec-proc  | <--|--- | render |<-------|    |
 *             +---------+      +-------------+    |    | process|        +----+
 *                             Acquire frame       |  	+--------+  Sync
 *                             Release frame                 |
 *                                                           | bind/unbind frame
 *    	 													 v	
 *                                                      +------------------+
 *                                                      | GST textsink lib |
 *                                                      +------------------+     
 * ]|   
 *
 * 
 * The following additional sink properties are added to support cross-process streaming:
 *
 * export-stream       : Allow video frames to be exported to a different
 *                       process. Disabled by default.
 * stream-token        : Read-only cross-process identifier for the stream.
 *
 * The rendering process is linked to an additional texture streaming utility 
 * library which provides the following functions for accessing buffers in the 
 * exported stream:
 *
 *|[
 * gboolean gst_texsink_open_stream (guint stream, EGLDisplay dpy);
 *]|
 *
 * Open an exported stream in the calling process. *stream* specifies the
 * stream handle which as given by the texture sink. *dpy* specifies the
 * EGL display used for rendering. Returns true on success and false on
 * failure.
 *
 *|[
 *  gboolean gst_texsink_bind_frame (guint stream, GLuint target, guint frame);
 *]|
 *
 * Bind a video frame in an exported stream to a texture target. *stream*
 * must be a handle of an opened stream. *target* must be
 * GL_TEXTURE_EXTERNAL_OES. *frame* gives the frame number to bind. If
 *  *frame* is -1, any previously bound frame is unbound.
 *|[
 * void gst_texsink_close_stream (guint stream);
 *]|
 *  Close a previously opened exported stream. *stream* specifies the
 *  stream handle.
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "meegovideotexture.h"

/*
 * enum and types
 */

enum
{
  /* emit signals */
  SIGNAL_FRAME_READY,
  LAST_SIGNAL
};


/*
 * defines and static global vars
 */
static guint meego_gst_video_texture_iface_signals[LAST_SIGNAL];

/*
 * Action signal function declarations
 */

static void meego_gst_video_texture_iface_base_init (MeegoGstVideoTextureClass * klass);
static void meego_gst_video_texture_iface_class_init (MeegoGstVideoTextureClass * klass);

GType
meego_gst_video_texture_get_type (void)
{
  static GType meego_gst_video_texture_type = 0;

  if (!meego_gst_video_texture_type) {
    static const GTypeInfo meego_gst_video_texture_info = {
      sizeof (MeegoGstVideoTextureClass),
      (GBaseInitFunc) meego_gst_video_texture_iface_base_init,        /* base_init */
      NULL,                     /* base_finalize */
      (GClassInitFunc) meego_gst_video_texture_iface_class_init,      /* class_init */
      NULL,                     /* class_finalize */
      NULL,                     /* class_data */
      0,
      0,                        /* n_preallocs */
      NULL,                     /* instance_init */
    };

    meego_gst_video_texture_type = g_type_register_static (G_TYPE_INTERFACE,
        "MeegoGstVideoTexture", &meego_gst_video_texture_info, 0);
    g_type_interface_add_prerequisite (meego_gst_video_texture_type,
        GST_TYPE_IMPLEMENTS_INTERFACE);
  }

  return meego_gst_video_texture_type;
}

static void
meego_gst_video_texture_iface_base_init (MeegoGstVideoTextureClass * klass)
{
  static gboolean initialized = FALSE;

  if (!initialized) {

  /**
   * MeegoGstVideoTexture::frame-ready:
   * 
   *  Triggered when a new frame is available from the sink
   *  element. frame specifies an opaque, non-zero
   *  identifier for the latest available frame. The
   *  application should call #meego_gst_video_texture_acquire_frame as a response
   *  to this signal.
   */

    meego_gst_video_texture_iface_signals[SIGNAL_FRAME_READY] =
        g_signal_new ("frame-ready",
        G_TYPE_FROM_CLASS (klass),
        G_SIGNAL_RUN_LAST,
        0, NULL, NULL,
        g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);

    initialized = TRUE;

  }

  klass->release_frame = NULL;
  klass->bind_frame = NULL;
  klass->acquire_frame = NULL;

}

static void
meego_gst_video_texture_iface_class_init (MeegoGstVideoTextureClass * klass)
{

  /**
   *  MeegoGstVideoTexture:egl-context
   *
   * EGL context
   * The application provided EGL display which will be used for rendering.
   * This property is used for proper memory management of textures in the sink
   * Application should update the egl-context in case of change.
   */
  g_object_interface_install_property (klass,
      g_param_spec_pointer ("egl-context", "Egl context",
          "Egl context created by the application", G_PARAM_READWRITE));
  /**
   *  MeegoGstVideoTexture:display
   *
   * EGL display
   * The application provided EGL display which will be used for rendering.
   */
  g_object_interface_install_property (klass,
      g_param_spec_pointer ("egl-display",
          "EGL display ",
          "The application provided EGL display which will be used for rendering.",
          G_PARAM_READWRITE));

  /**
   *  MeegoGstVideoTexture:export-stream
   *
   * Allow video frames to be exported to a different process.
   * 
   */
  g_object_interface_install_property (klass,
      g_param_spec_boolean ("export-stream",
          "Export Stream Token",
          "Allow video frames to be exported to a different process.",
          FALSE, G_PARAM_READWRITE));
  /**
   *  MeegoGstVideoTexture:stream-token
   *
   * 
   *  Read-only cross-process identifier for the stream.
   */
  g_object_interface_install_property (klass,
      g_param_spec_int ("stream-token",
          "Stream Token",
          "Read-only cross-process identifier for the stream.",
          0, G_MAXINT32, 0, G_PARAM_READABLE));


}


/**
 * meego_gst_video_texture_acquire_frame:
 * @vt: #MeegoGstVideoTexture interface of a #GstElement
 * @frame: identifies the video frame as given by the frame-ready signal.
 * 
 * Called by the application to lock a frame to be used in
 * graphics rendering. 
 *
 * Returns: %TRUE on success and %FALSE on failure.
 */
gboolean
meego_gst_video_texture_acquire_frame (MeegoGstVideoTexture * vt, gint frame)
{
  MeegoGstVideoTextureClass *iface = MEEGO_GST_VIDEO_TEXTURE_GET_CLASS (vt);

  if (iface->acquire_frame) {
    return iface->acquire_frame (vt, frame);
  }
  return FALSE;
}

/**
 * meego_gst_video_texture_bind_frame:
 * @vt: #MeegoGstVideoTexture interface of a #GstElement
 * @target: The only supported value is GL_TEXTURE_EXTERNAL_OES
 * @frame: identifies the video frame as given by the frame-ready signal.
 * 
 * Called by the application to bind a given frame to an
 * OpenGL (ES) texture target. The only supported value
 * for @target is GL_TEXTURE_EXTERNAL_OES as defined by
 * the GL_OES_EGL_image_external extension. @frame
 * identifies the video frame to bind, and a value of -1
 * has the effect of unbinding any previously bound frame
 * from @target.
 *
 * An OpenGL (ES) context must be current at the time of
 * this call.
 * The application can use glGetError() to get additional
 * information about a failure that occured during the
 * binding.
 *
 * Returns: %TRUE on success and %FALSE on failure.
 */
gboolean
meego_gst_video_texture_bind_frame (MeegoGstVideoTexture * vt, gint target, gint frame)
{
  MeegoGstVideoTextureClass *iface = MEEGO_GST_VIDEO_TEXTURE_GET_CLASS (vt);

  if (iface->acquire_frame) {
    return iface->bind_frame (vt, target, frame);
  }
  return FALSE;
}

/**
 * meego_gst_video_texture_release_frame:
 * @vt: #MeegoGstVideoTexture interface of a #GstElement
 * @frame: identifies the video frame as given by the frame-ready signal.
 * @sync: fence object used for synchronization
 * 
 * Called by the application to unlock a frame after it
 * has been used in graphics rendering. Before triggering
 * this action, the application is expected to use the
 * EGL_KHR_sync_fence extension to place a fence into the
 * GPU command stream after the rendering commands
 * accessing the video frame. This sync object should be
 * given as @sync, and the sink assumes ownership of that
 * sync object.
 *
 * If @sync is not 0, then the sink implementation
 * asynchronously waits until the specified one-time fence
 * object is triggered before releasing the image back to
 * the decoder. Otherwise, this function performs a
 * synchronous GPU flush (i.e. eglWaitClient()) to ensure
 * that the image not being read from.
 *
 */
void
meego_gst_video_texture_release_frame (MeegoGstVideoTexture * vt, gint frame,
    gpointer sync)
{
  MeegoGstVideoTextureClass *iface = MEEGO_GST_VIDEO_TEXTURE_GET_CLASS (vt);

  if (iface->release_frame) {
    return iface->release_frame (vt, frame, sync);
  }
  return;
}

/**
 * meego_gst_video_texture_frame_ready:
 * @vt: #MeegoGstVideoTexture interface of a #GstElement
 * @frame: identifies the video frame as given by the frame-ready signal.
 * 
 * Called by the elements implementing the #MeegoGstVideoTexture interface to 
 * trigger the signal emmision of the frame-ready signal 
 */
void
meego_gst_video_texture_frame_ready (MeegoGstVideoTexture * vt, gint frame)
{
  g_return_if_fail (MEEGO_GST_IS_VIDEO_TEXTURE (vt));

  g_signal_emit (G_OBJECT (vt),
      meego_gst_video_texture_iface_signals[SIGNAL_FRAME_READY], 0, frame);
}
